/*
 * Copyright (c) 2005 Brian Tarricone <bjt23@cornell.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License ONLY.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif

#include <X11/Xlib.h>
#include <glib.h>

#include "transd.h"

static Transd *__g_transd = NULL;
static Display *__g_dpy = NULL;
static gboolean running = FALSE;


static void
sighandler(int sig)
{
    switch(sig) {
        case SIGHUP:
        case SIGUSR1:
            transd_free(__g_transd);
            __g_transd = transd_new(__g_dpy);
            break;
    }
}

static void
x_main(Display *dpy,
       Transd *transd)
{
    while (running) {
        XEvent xev;
        XNextEvent(dpy, &xev);

        if (xev.type == MapNotify)
            transd_window_mapped(transd, &xev.xmap);
        else if (xev.type == PropertyNotify)
            transd_property_changed(transd, &xev.xproperty);
    }
}

int
main(int argc, char **argv)
{
    gint i;
    gboolean daemon_mode = TRUE;
    const char *display_name = NULL;
    
    for(i = 1; i < argc; i++) {
        if(!strcmp(argv[i], "--debug"))
            daemon_mode = FALSE;
        else if (!strncmp(argv[1], "--display=", 10))
            display_name = argv[1] + 10;
    }
    
    if(daemon_mode) {
        switch(fork()) {
            case -1:
                g_error("Unable to fork(): %s", strerror(errno));
                exit(errno);  /* NOTREACHED */
                break;
            
            case 0:
                /* we're the child, try to close all fds and reopen properly */
                {
                    gchar *xsession_errors;
                    
                    setsid();
                    for(i = getdtablesize() - 1; i >= 0; i--)
                        close(i);
                    
                    xsession_errors = g_build_filename(g_get_home_dir(),
                                                       ".xsession-errors",
                                                       NULL);
                    open("/dev/null", O_RDONLY);  /* stdin */
                    i = open(xsession_errors, O_CREAT|O_APPEND, 600);  /* stdout */
                    if(i >= 0) {
                        if (dup(i))  /* stderr */
                            g_printerr("Failed to dup() stdout for stderr\n");
                    }
                    g_free(xsession_errors);
                }
                break;
            
            default:
                /* we're the parent, bail */
                _exit(0);
        }
    }
    
    __g_dpy = XOpenDisplay(display_name);
    if (!__g_dpy) {
        g_printerr("Unable to open display %s\n", display_name ? display_name : ":0");
        return EXIT_FAILURE;
    }

    signal(SIGHUP, sighandler);
    signal(SIGUSR1, sighandler);
    
    __g_transd = transd_new(__g_dpy);
    
    running = TRUE;
    x_main(__g_dpy, __g_transd);
    
    transd_free(__g_transd);
    XCloseDisplay(__g_dpy);
    
    return EXIT_SUCCESS;
}
