/*
 * Copyright (c) 2005 Brian Tarricone <bjt23@cornell.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License ONLY.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#include <X11/Xlib.h>
#include <X11/Xatom.h>

#include <glib.h>
#include <libxfce4util/libxfce4util.h>

#include "transd.h"

#define _XA_OPACITY           "_NET_WM_WINDOW_OPACITY"
#define _XA_WM_CLASS          "WM_CLASS"
#define _XA_WM_NAME           "WM_NAME"
#define _XA_NET_WM_NAME       "_NET_WM_NAME"
#define _XA_ROLE              "WM_WINDOW_ROLE"
#define TRANSD_RULES_CONFIG   "transd/rules.rc"

struct _Transd
{
    Display *dpy;
    GHashTable *rules;  /* (gchar *) -> (TransdClassRule *) */
};

typedef struct
{
    gint all_window_opacity;
    GHashTable *roles;  /* (gchar *) -> (gint) */
} TransdClassRule;


static int (*last_x_error_handler)(Display *, XErrorEvent *) = NULL;
static int last_x_error = 0;

static int
x_error_handler(Display *dpy,
                XErrorEvent *xev)
{
    last_x_error = xev->error_code;
    return 0;
}

static void
trap_x_errors(void)
{
    g_assert(last_x_error_handler == NULL);
    last_x_error_handler = XSetErrorHandler(x_error_handler);
}

static int
untrap_x_errors(Display *dpy)
{
    int ret = last_x_error;
    XSync(dpy, False);
    XSetErrorHandler(last_x_error_handler);
    last_x_error_handler = NULL;
    last_x_error = 0;
    return ret;
}

static void
transd_class_rule_free(TransdClassRule *class_rule)
{
    g_return_if_fail(class_rule);
    
    g_hash_table_destroy(class_rule->roles);
    g_free(class_rule);
}

static gchar *
get_prop(Display *dpy,
         Window window,
         const char *prop_name,
         Atom prop_type,
         size_t *len)
{
    gchar *prop_value = NULL;
    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *prop_return = NULL;
    
    g_return_val_if_fail(window != None, NULL);

    trap_x_errors();

    if (Success == XGetWindowProperty(dpy, window,
                                      XInternAtom(dpy, prop_name, False),
                                      0L, 8192, False, prop_type, &actual_type,
                                      &actual_format, &nitems, &bytes_after,
                                      &prop_return))
    {
        if (actual_format == 8 && actual_type == prop_type) {
            prop_value = g_memdup(prop_return, nitems);
            if (len)
                *len = nitems;
        }
        
        if (prop_return)
            XFree(prop_return);
    }

    untrap_x_errors(dpy);
    
    return prop_value;
}

static gchar *
get_string_prop(Display *dpy,
                Window window,
                const char *prop_name,
                size_t *len)
{
    return get_prop(dpy, window, prop_name, XA_STRING, len);
}

static gchar *
get_utf8_string_prop(Display *dpy,
                     Window window,
                     const char *prop_name,
                     size_t *len)
{
    return get_prop(dpy, window, prop_name, XInternAtom(dpy, "UTF8_STRING", False), len);
}

static gchar *
get_class_name(Display *dpy,
               Window window)
{
    gchar *class_name = NULL;
    size_t len = 0;
    gchar *wm_class = get_string_prop(dpy, window, _XA_WM_CLASS, &len);
    if (wm_class) {
        size_t instance_len = strlen(wm_class);
        if (len > instance_len + 1)
            class_name = g_strndup(wm_class + instance_len + 1, len - instance_len + 1);
        g_free(wm_class);
    }

    return class_name;
}

static gboolean
transd_read_config(Transd *transd, const gchar *filename)
{
    XfceRc *rcfile;
    gchar **groups = NULL;
    gchar **entries = NULL;
    gint i, j;
    guint32 value;
    TransdClassRule *class_rule;
    
    TRACE("entering, filename=%s", filename);
    
    rcfile = xfce_rc_simple_open(filename, TRUE);
    if(!rcfile)
        return FALSE;
    
    if(!xfce_rc_has_group(rcfile, "transd-global-config")) {
        g_printerr("%s\n", _("Transd now uses a new configuration file format.  Please see the README for details.  Transd will now exit."));
        exit(1);
    }
    
    xfce_rc_set_group(rcfile, "transd-global-config");
    /* load global data that i haven't thought of yet */
    
    groups = xfce_rc_get_groups(rcfile);
    for(i = 0; groups[i]; ++i) {
        if(!strcmp(groups[i], "transd-global-config")) {
            g_free(groups[i]);
            continue;
        }
        
        entries = xfce_rc_get_entries(rcfile, groups[i]);
        if(!entries)
            continue;
        xfce_rc_set_group(rcfile, groups[i]);
        
        class_rule = g_new(TransdClassRule, 1);
        class_rule->all_window_opacity = xfce_rc_read_int_entry(rcfile,
                                                                "transd-ALL",
                                                                100);
        class_rule->roles = g_hash_table_new_full(g_str_hash, g_str_equal,
                                                  (GDestroyNotify)g_free, NULL);
        g_hash_table_replace(transd->rules, groups[i], class_rule);
        
        for(j = 0; entries[j]; ++j) {
            if(!strcmp(entries[j], "transd-ALL")) {
                g_free(entries[j]);
                continue;
            }
            
            value = xfce_rc_read_int_entry(rcfile, entries[j], -1);
            if(value <= 100) {
                g_hash_table_replace(class_rule->roles, entries[j],
                                     GINT_TO_POINTER(value+1));
                DBG("  added config class '%s', role '%s' => 0x%x", groups[i],
                    entries[j], value);
            }
        }
        g_free(entries);  /* NOT g_strfreev() */
    }
    
    g_free(groups);  /* NOT g_strfreev() */
    xfce_rc_close(rcfile);
    
    return TRUE;
}

static void
transd_apply_rule(Transd *transd, Window window)
{
    gchar *class_;
    gchar *role;
    gint percentage = 0;
    guint32 opacity;
    TransdClassRule *class_rule;
    
    TRACE("entering");
    
    g_return_if_fail(transd && window != None);

    class_ = get_class_name(transd->dpy, window);
    if (!class_) {
        class_ = get_string_prop(transd->dpy, window, _XA_WM_CLASS, NULL);
        if (!class_) {
            class_ = get_utf8_string_prop(transd->dpy, window, _XA_NET_WM_NAME, NULL);
            if (!class_) {
                class_ = get_string_prop(transd->dpy, window, _XA_WM_NAME, NULL);
                if (!class_)
                    return;
            }
        }
    }
    DBG("got class '%s'", class_);

    /* percentage is actually percentage+1 since a value of 0% would be
     * interpreted as "not in hash table" */
    class_rule = g_hash_table_lookup(transd->rules, class_);
    g_free(class_);
    if(!class_rule)
        return;
    
    role = get_string_prop(transd->dpy, window, _XA_ROLE, NULL);
    DBG("got role '%s'", role);
    if(role)
        percentage = GPOINTER_TO_INT(g_hash_table_lookup(class_rule->roles, role));
    if(percentage == 0) {
        if(class_rule->all_window_opacity == 100) {
            g_free(role);
            return;
        } else
            percentage = class_rule->all_window_opacity;
    }
    
    opacity = (guint32)(((gdouble)(percentage - 1)) / 100 * 0xffffffff);
    
    DBG("applying opacity 0x%x to window %d", opacity, (guint32)window);
    
    trap_x_errors();
    XChangeProperty(transd->dpy, window,
                    XInternAtom(transd->dpy, _XA_OPACITY, FALSE),
                    XA_CARDINAL, 32, PropModeReplace,
                    (unsigned char *)&opacity, 1L);
    untrap_x_errors(transd->dpy);
    
    g_free(role);
}

    
Transd *
transd_new(Display *dpy)
{
    Transd *transd;
    gchar *config_system, *config_user;
    unsigned int i, n_screens;
    
    transd = g_new0(Transd, 1);
    transd->dpy = dpy;
    transd->rules = g_hash_table_new_full(g_str_hash, g_str_equal,
                                          (GDestroyNotify)g_free,
                                          (GDestroyNotify)transd_class_rule_free);
    
    config_system = xfce_resource_lookup(XFCE_RESOURCE_CONFIG,
                                         TRANSD_RULES_CONFIG);
    config_user = xfce_resource_save_location(XFCE_RESOURCE_CONFIG,
                                              TRANSD_RULES_CONFIG,
                                              FALSE);
    
    if(config_system && config_user && !strcmp(config_system, config_user))
        transd_read_config(transd, config_user);
    else if(config_system
            && (!config_user || !g_file_test(config_user, G_FILE_TEST_EXISTS)))
    {
        FILE *fp;
        
        if(!config_user)
            config_user = xfce_resource_save_location(XFCE_RESOURCE_CONFIG,
                                                      TRANSD_RULES_CONFIG,
                                                      TRUE);
        fp = fopen(config_user, "w");
        if(fp) {
            gchar *sysconfig_contents = NULL;
            gsize length = 0;
            GError *error = NULL;
            
            if(g_file_get_contents(config_system,
                                   &sysconfig_contents,
                                   &length,
                                   &error))
            {
                if (fwrite(sysconfig_contents, 1, length, fp) != length)
                    g_printerr("Failed to write full config file: %s\n", strerror(errno));
                fflush(fp);
                transd_read_config(transd, config_user);
            } else {
                g_critical("Unable to read systemwide config file: %s",
                           error->message);
                g_error_free(error);
            }
            fclose(fp);
        } else {
            g_critical("Unable to create per-user config file: %s",
                       strerror(errno));
            transd_read_config(transd, config_system);
        }
    } else if(config_system && config_user) {
        transd_read_config(transd, config_system);
        transd_read_config(transd, config_user);
    }
    
    g_free(config_system);
    g_free(config_user);

    n_screens = ScreenCount(dpy);
    for (i = 0; i < n_screens; ++i) {
        Window root = RootWindow(dpy, i);
        Window dummy;
        Window *children = NULL;
        unsigned int n_children = 0;

        trap_x_errors();
        XSelectInput(dpy, root, SubstructureNotifyMask);
        XQueryTree(dpy, root, &dummy, &dummy, &children, &n_children);
        untrap_x_errors(dpy);

        for (i = 0; i < n_children; ++i) {
            XMapEvent xev = {
                .type = MapNotify,
                .serial = 0,
                .send_event = False,
                .display = dpy,
                .event = root,
                .window = children[i],
                .override_redirect = False,
            };

            transd_window_mapped(transd, &xev);
        }

        if (children)
            XFree(children);
    }
    
    return transd;    
}

void
transd_window_mapped(Transd *transd,
                     XMapEvent *xmap)
{
    Window dummy;
    Window *children = NULL;
    unsigned int i, n_children = 0;

    XPropertyEvent xprop = {
        .type = PropertyNotify,
        .serial = 0,
        .send_event = False,
        .display = transd->dpy,
        .window = xmap->window,
        .atom = XInternAtom(transd->dpy, _XA_WM_CLASS, False),
        .state = PropertyNewValue,
    };

    trap_x_errors();
    XSelectInput(transd->dpy, xmap->window, PropertyChangeMask | SubstructureNotifyMask);
    untrap_x_errors(transd->dpy);
    transd_property_changed(transd, &xprop);

    trap_x_errors();
    XQueryTree(transd->dpy, xmap->window, &dummy, &dummy, &children, &n_children);
    untrap_x_errors(transd->dpy);
    for (i = 0; i < n_children; ++i) {
        trap_x_errors();
        XSelectInput(transd->dpy, children[i], PropertyChangeMask | SubstructureNotifyMask);
        untrap_x_errors(transd->dpy);

        xprop.window = children[i];
        transd_property_changed(transd, &xprop);
    }
    if (children)
        XFree(children);

}

void
transd_property_changed(Transd *transd,
                        XPropertyEvent *xprop)
{
    if (xprop->atom == XInternAtom(transd->dpy, "WM_CLASS", False) ||
        xprop->atom == XInternAtom(transd->dpy, "WM_WINDOW_ROLE", False) ||
        xprop->atom == XInternAtom(transd->dpy, "WM_NAME", False) ||
        xprop->atom == XInternAtom(transd->dpy, "_NET_WM_NAME", False))
    {
        transd_apply_rule(transd, xprop->window);
    }
}

void
transd_free(Transd *transd)
{
    g_return_if_fail(transd);

    g_hash_table_destroy(transd->rules);
    
    g_free(transd);
}
